package no.ntnu.mycbr.rest;


import com.sun.org.apache.regexp.internal.RE;
import de.dfki.mycbr.core.DefaultCaseBase;
import de.dfki.mycbr.core.ICaseBase;
import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.*;
import de.dfki.mycbr.core.model.*;
import de.dfki.mycbr.core.retrieval.Retrieval;
import de.dfki.mycbr.core.similarity.AdvancedIntegerFct;
import de.dfki.mycbr.core.similarity.IntegerFct;
import de.dfki.mycbr.core.similarity.Similarity;
import de.dfki.mycbr.core.similarity.SymbolFct;

import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.util.Pair;
import io.swagger.models.ComposedModel;
import io.swagger.models.auth.In;

import java.text.ParseException;
import java.util.*;

/**
 * Created by talep on 24.04.2017.
 */
public class AdaptationRules {

    private Project project;
    private Concept caseConcept;
    private Concept exerciseConcept;
    private ICaseBase case_cb;
    public static DefaultCaseBase exercise_cb;
    private ArrayList<String> intAdaptationAttributes = new ArrayList<>();
    private ArrayList<String> advancedIntAdaptationAttributes = new ArrayList<>();
    HashMap<Attribute, Attribute> attributeChange = new HashMap<>();
    int exerciseNr = 50;
    int caseNr = 15;


    public AdaptationRules() {
        this.project = App.getProject();
        this.caseConcept = project.getConceptByID("Case");
        this.exerciseConcept = project.getConceptByID("Exercise");
        this.case_cb = project.getCB("Case");
        this.exercise_cb = (DefaultCaseBase) project.getCaseBases().get("Exercise");

        //TODO: Read this in from file, add weights as well
        intAdaptationAttributes.add("BIPPQ_7");
        intAdaptationAttributes.add("Overall_health");
        intAdaptationAttributes.add("RMDQ");
        intAdaptationAttributes.add("Pain_lastweek_avg");
        intAdaptationAttributes.add("Pain_lastweek_max");
        intAdaptationAttributes.add("Pain_Self_Efficacy");

        advancedIntAdaptationAttributes.add("FABQ_PhysicalActivity");

    }

    public Case adaptCase(String similarCaseId, HashMap<String,Object> newCaseQueryInfo) {
        Instance similarCase = project.getInstance(similarCaseId);
        Instance newCase = createInstanceFromQuery(caseConcept, newCaseQueryInfo);

        double adaptationScore = compareCases(newCase, similarCase);

        List<Instance> exercises = retrieveSolution(similarCase);
        List<Instance> newExercises = solutionAdaptation(exercises, adaptationScore);


//        AttributeDesc attDesc = caseConcept.getAttributeDesc("Exercise_list");
//        System.out.println("Before ad: " + newCase.getAttributes().get(attDesc));
//
//        ConceptDesc conceptDesc = (ConceptDesc) attDesc;
//        LinkedList<Attribute> llAtts = new LinkedList<>();
//        Iterator it = newExercises.iterator();
//        while (it.hasNext()){
//            Object exercise = it.next();
//            try {
//                llAtts.add(conceptDesc.getAttribute(exercise));
//            } catch (ParseException e) {
//                e.printStackTrace();
//            }
//        }
//        MultipleAttribute<ConceptDesc> muliSymbol = new MultipleAttribute<>(conceptDesc, llAtts);
//        newCase.addAttribute(attDesc, muliSymbol);
//
//        System.out.println("After ad: " + newCase.getAttributes().get(attDesc));
        
        return null;
    }

    private Instance createInstanceFromQuery(Concept concept, HashMap<String, Object> queryContent){
        Instance newCase = new Instance(concept, ("Case #" + caseNr));
        caseNr += 1;

        try {
            for (Map.Entry<String, Object> att : queryContent.entrySet()) {
                String name = att.getKey();
                AttributeDesc attdesc = concept.getAttributeDesc(name);
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("FloatDesc")){
                    newCase.addAttribute(attdesc, Float.parseFloat(att.getValue().toString()));
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("IntegerDesc")){
                    newCase.addAttribute(attdesc, Integer.parseInt(att.getValue().toString()));
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("DoubleDesc")){
                    newCase.addAttribute(attdesc, Double.parseDouble(att.getValue().toString()));
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("SymbolDesc")){
                    SymbolDesc aSymbolAtt = (SymbolDesc) attdesc;
                    if (!aSymbolAtt.isMultiple()) {
                        newCase.addAttribute(attdesc, (String) att.getValue());
                    }
                    else {
                        LinkedList<Attribute> llAtts = new LinkedList<Attribute>();
                        StringTokenizer st = new StringTokenizer((String) att.getValue(), ",");
                        while (st.hasMoreElements()) {
                            String symbolName = st.nextElement().toString().trim();
                            llAtts.add(aSymbolAtt.getAttribute(symbolName));
                        }

                        MultipleAttribute<SymbolDesc> muliSymbol = new MultipleAttribute<SymbolDesc>(aSymbolAtt, llAtts);
                        newCase.addAttribute(attdesc, muliSymbol);
                    }
                }
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        return newCase;
    }

    private double compareCases(Instance similarCase, Instance newCase){

        double changeScore = .0;

        for (String attr : intAdaptationAttributes) {
            AttributeDesc attDesc = caseConcept.getAttributeDesc(attr);
            IntegerFct f = (IntegerFct) caseConcept.getActiveAmalgamFct().getActiveFct(attDesc);
            try {
                Attribute simCaseAtt = similarCase.getAttForDesc(attDesc);
                Attribute newCaseAtt = newCase.getAttForDesc(attDesc);
                if (Integer.parseInt(simCaseAtt.getValueAsString()) > Integer.parseInt(newCaseAtt.getValueAsString())){
                    changeScore += (f.calculateSimilarity(simCaseAtt, newCaseAtt).getValue() - 1.);
                } else if (Integer.parseInt(simCaseAtt.getValueAsString()) < Integer.parseInt(newCaseAtt.getValueAsString())) {
                    changeScore += (1. - f.calculateSimilarity(simCaseAtt, newCaseAtt).getValue());
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        for (String attr : advancedIntAdaptationAttributes) {
            AttributeDesc attDesc = caseConcept.getAttributeDesc(attr);
            AdvancedIntegerFct f = (AdvancedIntegerFct) caseConcept.getActiveAmalgamFct().getActiveFct(attDesc);
            try {
                Attribute simCaseAtt = similarCase.getAttForDesc(attDesc);
                Attribute newCaseAtt = newCase.getAttForDesc(attDesc);
                if (Integer.parseInt(simCaseAtt.getValueAsString()) > Integer.parseInt(newCaseAtt.getValueAsString())){
                    changeScore += (f.calculateSimilarity(simCaseAtt, newCaseAtt).getValue() - 1.);
                } else if (Integer.parseInt(simCaseAtt.getValueAsString()) < Integer.parseInt(newCaseAtt.getValueAsString())) {
                    changeScore += (1. - f.calculateSimilarity(simCaseAtt, newCaseAtt).getValue());
                }
                System.out.println("Score for endring er nå: " + changeScore);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        System.out.println("Normalized score: " + normalize(changeScore));
        return normalize(changeScore);
    }

    // TODO: move duplicate code here
    private double calculateScore(){

        return 0;
    }

    private double normalize(double score){
        double max_x = .0 + this.intAdaptationAttributes.size() + this.advancedIntAdaptationAttributes.size();
        double min_x = -(max_x);

        return (2.*((score - min_x)/(max_x-min_x)))-1.;
    }
    
    private List<Instance> retrieveSolution(Instance instance){
        AttributeDesc exerciseListAttr = caseConcept.getAttributeDesc("Exercise_list");
        Attribute exerciseList = instance.getAttributes().get(exerciseListAttr);

        List<Instance> exercises = new ArrayList<>();

        for(String exerciseId : exerciseList.getValueAsString().split(";")) {
            exercises.add(exerciseConcept.getInstance(exerciseId));
        }


        return exercises;
    }

    private List<Instance> solutionAdaptation(List<Instance> exercises, double changeScore) {

        List<Instance> adaptedSolution = new ArrayList<>(exercises);
        Collections.copy(adaptedSolution, exercises);
        AttributeDesc repDesc = exerciseConcept.getAttributeDesc("Repetitions");
        AttributeDesc levelDesc = exerciseConcept.getAttributeDesc("Level");
        AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");
        AttributeDesc setDesc = exerciseConcept.getAttributeDesc("Set");

        int changedExercises = 0;

        while (Math.abs(changeScore)*100 > changedExercises) {
            try {

                // Sort list, get 0 element, change, repeat
                if (changeScore > 0){
                    adaptedSolution = sortAscending(adaptedSolution);
                } else {
                    adaptedSolution = sortDescending(adaptedSolution);
                }

                Instance exerciseToChange = adaptedSolution.get(0);
                System.out.println("Exercise to change " + exerciseToChange.getName());
                Instance changedExercise = changeExercise(exerciseToChange, changeScore > 0);
                System.out.println("Exercise changed " + changedExercise.getName());
                Instance newExercise = exerciseInCb(changedExercise);
                System.out.println("New exercise " + newExercise.getName());
                if(newExercise.getName().equals(changedExercise.getName())){
                    System.out.println("New exercise: " + newExercise.getName() + " " + newExercise.getAttForDesc(descriptionDesc) + ", reps: " + newExercise.getAttForDesc(repDesc) + ", set: " + newExercise.getAttForDesc(setDesc).getValueAsString() + ", level: " + newExercise.getAttForDesc(levelDesc).getValueAsString());
                    exercise_cb.addCase(changedExercise);
                }
                System.out.println("Before one change: " + adaptedSolution);
                adaptedSolution.remove(exerciseToChange);
                adaptedSolution.add(newExercise);
                System.out.println("After one change: " + adaptedSolution);
                System.out.println("----------------------------------------------");

            } catch (Exception e) {
                e.printStackTrace();
            }
            changedExercises += 1;
        }

        System.out.println("Original solution: " + exercises.toString());
        System.out.println("Adapted solution: " + adaptedSolution.toString());

        return adaptedSolution;
    }

    private List<Instance> sortWithoutFlex(List<Instance> exercises){
        Collections.sort(exercises, new Comparator<Instance>() {
            AttributeDesc repDesc = exerciseConcept.getAttributeDesc("Repetitions");
            AttributeDesc levelDesc = exerciseConcept.getAttributeDesc("Level");
            AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");
            @Override
            public int compare(Instance e1, Instance e2) {
                int num = 0;
                if(e1.getAttForDesc(repDesc).getValueAsString() != "_unknown_" && e2.getAttForDesc(repDesc).getValueAsString() != "_unknown_"){
                    num = (Integer.parseInt(e1.getAttForDesc(repDesc).getValueAsString()) - Integer.parseInt(e2.getAttForDesc(repDesc).getValueAsString()));
                }
                if(num == 0){
                    num = (Integer.parseInt(e1.getAttForDesc(levelDesc).getValueAsString()) - Integer.parseInt(e2.getAttForDesc(levelDesc).getValueAsString()));
                }
                if(e1.getAttForDesc(descriptionDesc).getValueAsString().equals("Flexibility")) {
                    num = 1;
                }
                return num;
            }
        });
        return exercises;
    }

    private List<Instance> sortAscending(List<Instance> exercises){
        AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");

        List<Instance> flex = new ArrayList<>();
        List<Instance> sortExercise = new ArrayList<>();
        for (Instance exercise:exercises) {
            if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Flexibility")){
                flex.add(exercise);
            } else {
                sortExercise.add(exercise);
            }
        }

        sortExercise = sortWithoutFlex(sortExercise);
        sortExercise.addAll(flex);
        System.out.println("SORTED??: " + sortExercise);

        return sortExercise;
    }

    private List<Instance> sortDescending(List<Instance> exercises){
        AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");

        List<Instance> flex = new ArrayList<>();
        List<Instance> sortExercise = new ArrayList<>();
        for (Instance exercise:exercises) {
            if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Flexibility")){
                flex.add(exercise);
            } else {
                sortExercise.add(exercise);
            }
        }

        sortExercise = sortWithoutFlex(sortExercise);
        Collections.reverse(sortExercise);
        sortExercise.addAll(flex);
        System.out.println("SORTED??: " + sortExercise);

        return sortExercise;
    }


    private Instance changeExercise(Instance exercise, Boolean increase) throws Exception {

        AttributeDesc repDesc = exerciseConcept.getAttributeDesc("Repetitions");
        AttributeDesc levelDesc = exerciseConcept.getAttributeDesc("Level");

        Instance copied = exerciseConcept.copyInstance(exercise, ("Exercise #" + exerciseNr));
        exerciseNr += 1;

        Attribute rep = copied.getAttForDesc(repDesc);

        int oldRep = Integer.parseInt(rep.getValueAsString());
        int oldLevel = Integer.parseInt(copied.getAttForDesc(levelDesc).getValueAsString());
        String newRep = "0";

        if(increase && canBeIncreased(exercise)){
            System.out.println("-------------OPPJUSTER---------------");
            // Finn ut hvordan reps skal endres
            if(oldRep == 8){
                newRep = "10";
            } else if (oldRep == 10){
                newRep = "12";
            } else if (oldRep == 12){
                newRep = "15";
            } else if (oldRep == 15 && canBeIncreased(exercise)){
                newRep = "8";
            }

            // Sett nye reps
            if(newRep != "0") {
                copied.addAttribute(repDesc.getName(), repDesc.getAttribute(newRep));

                // Hvis rep 15->8 øk ex level
                if(newRep == "8") {
                    String newLevel = Integer.toString(oldLevel+1);
                    copied.addAttribute(levelDesc, newLevel);
                }
                copied.setAdapted(true);
            }

        } else if (!increase){
            System.out.println("-------------NEDJUSTER---------------");

            if(oldRep == 8 && oldLevel > 1){
                newRep = "15";
            } else if (oldRep == 10){
                newRep = "8";
            } else if (oldRep == 12){
                newRep = "10";
            } else if (oldRep == 15){
                newRep = "12";
            }

            // Sett nye reps
            if(newRep != "0") {
                copied.addAttribute(repDesc.getName(), repDesc.getAttribute(newRep));

                // Hvis reps 8->15 så mink exercise level
                if(newRep == "15") {
                    String newLevel = Integer.toString(oldLevel-1);
                    copied.addAttribute(levelDesc, newLevel);
                }

                copied.setAdapted(true);
            }

        }
        return copied;
    }

    // TODO: Les inn maxLevel fra fil
    private boolean canBeIncreased(Instance exercise){
        AttributeDesc levelDesc = exerciseConcept.getAttributeDesc("Level");
        AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");
        AttributeDesc repDesc = exerciseConcept.getAttributeDesc("Repetitions");

        if(Integer.parseInt(exercise.getAttForDesc(repDesc).getValueAsString()) < 15){
            return true;
        }

        if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the abdominal muscles")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 6);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the back extensors")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 4);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the gluts and back extensors")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 4);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the oblique abdominals and rotator muscle in the back")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 6);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength “Strong in mid-position” ")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 4);
        }

        return false;
    }

    private Instance exerciseInCb(Instance exercise){
        Retrieval r = new Retrieval(exerciseConcept, exercise_cb);

        try {
            Instance query = r.getQueryInstance();

            for (Map.Entry<AttributeDesc, Attribute> att : exercise.getAttributes().entrySet()) {
                AttributeDesc attdesc = att.getKey();
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("FloatDesc")){
                    query.addAttribute(attdesc, att.getValue());
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("IntegerDesc")){
                    query.addAttribute(attdesc, att.getValue());
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("DoubleDesc")){
                    query.addAttribute(attdesc, att.getValue());
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("SymbolDesc")){
                    query.addAttribute(attdesc, att.getValue());
                }
            }

            r.setK(1);
            r.setRetrievalMethod(Retrieval.RetrievalMethod.RETRIEVE_K_SORTED);

            r.start();

            List<Pair<Instance, Similarity>> results = r.getResult();


            for (Pair<Instance, Similarity> result : results) {
                double sim = result.getSecond().getValue();
                if(sim == 1){
                    return result.getFirst();
                }
            }
            query.reset();

        }
        catch (Exception e) {
            e.printStackTrace();
        }

        return exercise;
    }

}
