package no.ntnu.mycbr.rest;

import de.dfki.mycbr.core.casebase.Instance;
import io.swagger.annotations.*;
import no.ntnu.mycbr.CBREngine;
import org.springframework.web.bind.annotation.*;

import java.util.*;

/**
 * Created by kerstin on 05/08/16.
 */
@RestController
public class CBRController {

    @ApiOperation(value = "getConcept", nickname = "getConcept")
    @RequestMapping(method = RequestMethod.GET, path="/concepts", produces = "application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = ConceptName.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public ConceptName getConcept() {
        return new ConceptName();
    }


    @ApiOperation(value = "getSimilarCases", nickname = "getSimilarCases")
    @RequestMapping(method = RequestMethod.GET, path="/retrieval", produces = "application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = Query.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public Query getSimilarCases(@RequestParam(value="casebase", defaultValue="Example_CB") String casebase,
                          @RequestParam(value="concept name", defaultValue="Case") String concept,
                          @RequestParam(value="Symbol attribute name", defaultValue="Dem_Gender") String attribute,
                          @RequestParam(value="value", defaultValue="male") String value) {
        return new Query(casebase, concept, attribute, value);
    }

    @ApiOperation(value = "getSimilarCases", nickname = "getSimilarCases")
    @RequestMapping(method = RequestMethod.POST, path="/retrieval", produces = "application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = Query.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public Query getSimilarCases(@RequestParam(value="casebase", defaultValue="Example_CB") String casebase,
                                 @RequestParam(value="concept name", defaultValue="Case") String concept,
                                 @RequestParam(required = false, value="no of returned cases",defaultValue = "-1") int k,
                                 @RequestBody(required = true)  HashMap<String, Object> queryContent) {
        return new Query(casebase, concept, queryContent, k);
    }

    @ApiOperation(value = "getSimilarCasesByID", nickname = "getSimilarCasesByID")
    @RequestMapping(method = RequestMethod.GET, path="/retrievalByID", produces = "application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = Query.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public Query getSimilarCases(@RequestParam(value="casebase", defaultValue="Example_CB") String casebase,
                                     @RequestParam(value="concept", defaultValue="Case") String concept,
                                     @RequestParam(value="caseID", defaultValue="Case #0") String caseID) {
        return new Query(casebase, concept, caseID);
    }


    @ApiOperation(value = "getCase", nickname = "getCase")
    @RequestMapping(method = RequestMethod.GET, value = "/case", headers="Accept=application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = Case.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public Case getCase(@RequestParam(value="caseID", defaultValue="Case #0") String caseID) {
        return new Case(caseID);
    }


    @ApiOperation(value = "getCaseBases", nickname = "getCaseBases")
    @RequestMapping(method = RequestMethod.GET, path="/casebase", produces = "application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = CaseBases.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public CaseBases getCaseBases() {
        return new CaseBases();
    }

    @ApiOperation(value = "getAttributes", nickname = "getAttributes")
    @RequestMapping(method = RequestMethod.GET, value = "/attributes", headers="Accept=application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = ApiResponse.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public Attribute getAttributes(@RequestParam(value="concept name", defaultValue="Case") String concept) {
        return new Attribute(concept);
    }

    @ApiOperation(value = "getValueRange", nickname = "getValueRange")
    @RequestMapping(method = RequestMethod.GET, value = "/values", headers="Accept=application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = ValueRange.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public ValueRange getValueRange(@RequestParam(value="concept name", defaultValue="Case") String concept,
                                    @RequestParam(value="attribute name", defaultValue="Pain_Medication") String attributeName) {
        return new ValueRange(concept, attributeName);
    }

    @ApiOperation(value = "getGaGeneratedCase", nickname = "getGaCase")
    @RequestMapping(method = RequestMethod.POST, value = "/gacase", headers="Accept=application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = Case.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public Case getGaGeneratedCase(@RequestParam(value="casebase", defaultValue="Example_CB") String casebase,
                                   @RequestParam(value="concept name", defaultValue="Case") String concept,
                                   @RequestParam(required = false, value="no of returned cases",defaultValue = "-1") int k,
                                   @RequestBody(required = true)  HashMap<String, Object> queryContent) {
        Query query = new Query(casebase, concept, queryContent, k);
        List<String> caseNames = new ArrayList<>();
        for ( String key : query.getSimilarCases().keySet() ) {
            caseNames.add(key);
        }
        GA ga = new GA();
        System.out.println("Created GA instance");
        Case gaCaze = ga.crossover(caseNames.get(0), caseNames.get(1), queryContent);
        return gaCaze;
    }

    @ApiOperation(value = "getAdaptedCase", nickname = "getAdaptedCase")
    @RequestMapping(method = RequestMethod.POST, value = "/getAdaptedCase", headers="Accept=application/json")
    @ApiResponses(value = {
            @ApiResponse(code = 200, message = "Success", response = Case.class),
            @ApiResponse(code = 401, message = "Unauthorized"),
            @ApiResponse(code = 403, message = "Forbidden"),
            @ApiResponse(code = 404, message = "Not Found"),
            @ApiResponse(code = 500, message = "Failure")})
    public String getAdaptedCase(@RequestParam(value="casebase", defaultValue="Example_CB") String casebase,
            @RequestParam(value="concept name", defaultValue="Case") String concept,
            @RequestBody(required = true)  HashMap<String, Object> queryContent) {

        Query query = new Query(casebase, concept, queryContent, 1);
        String caseName = query.getSimilarCases().keySet().iterator().next();
        AdaptationRules adaptationRules = new AdaptationRules();
        adaptationRules.adaptCase(caseName, queryContent);
        return null;
    }

}
