package no.ntnu.mycbr.rest;

import de.dfki.mycbr.core.DefaultCaseBase;
import de.dfki.mycbr.core.ICaseBase;
import de.dfki.mycbr.core.Project;
import de.dfki.mycbr.core.casebase.*;
import de.dfki.mycbr.core.casebase.Attribute;
import de.dfki.mycbr.core.model.AttributeDesc;
import de.dfki.mycbr.core.model.Concept;
import de.dfki.mycbr.core.model.SymbolDesc;
import no.ntnu.mycbr.CBREngine;
import org.springframework.context.annotation.Description;

import javax.validation.constraints.Null;
import java.text.ParseException;
import java.util.*;

/**
 * Created by talep on 13.01.2017.
 */
public class GA {

    private double mutationRate = 0.015;
    private String[] crossoverAttributes = {"Exercise_1", "Exercise_2", "Exercise_3" };
    private Concept myConcept;
    private Project project;
    private Concept exerciseConcept;
    public static DefaultCaseBase exercise_cb;
    int caseNr = 15;
    int exerciseNr = 50;

    public GA() {
        this.project = App.getProject();
        this.myConcept = project.getConceptByID("Case");
        this.exerciseConcept = project.getConceptByID("Exercise");
        this.exercise_cb = (DefaultCaseBase) project.getCaseBases().get("Exercise");
    }


    /*
    Crossover function.
    Input: Two cbrCases, content from query case
    Output: New case created from querycontent with exerciselist created from cbrCases
    * */
    public Case crossover(String cbrCase1, String cbrCase2, HashMap<String, Object> queryContent) {


        //Hent ut exerciselist på to næreste cases
        String exerciseList1 = new Case(cbrCase1).getCase().get("Exercise_list");
        String exerciseList2 = new Case(cbrCase2).getCase().get("Exercise_list");




        //Lage nye exerciselist basert på to gamle, bruk den ene om den andre ikke er satt
        String newExerciseList;
        if(exerciseList1 == "_unknown_" && exerciseList2 == "_unknown_" ) {
            newExerciseList = ""; //TODO: sette painrelief exercises
        }
        else if(exerciseList1 == "_unknown_") {
            newExerciseList = exerciseList2;
        }
        else if(exerciseList2 == "_unknown_") {
            newExerciseList = exerciseList1;
        }
        else {
            System.out.println("Crossover");
            newExerciseList = crossover(exerciseList1, exerciseList2);

        }


        //Finn ut om du skal mutere

        Random generator = new Random();
        int randomNum = generator.nextInt(1000);
        if(randomNum < 16) {
            newExerciseList = mutation(newExerciseList);
        }


        // Legg til exerciselist i queryinnhold. Får lagt til, men legges ikke på instansen
        //System.out.println("new exercise list: " + newExerciseList);
        //System.out.println("Old querycontent: " + queryContent.entrySet());
        //queryContent.put("Exercise_List", newExerciseList);
        //System.out.println("New qc: " + queryContent.entrySet());


        // Lage ny instans med queryinnhold
        Instance instance = createInstanceFromQuery(myConcept ,queryContent);

        System.out.println(newExerciseList);
        //Lagre ny exerciselist på instansen TODO



        //Lagre ny instans i CB og returner ny case

//        cb.addCase(caze3);
//        Instance addedCase = cb.containsCase(caze3.getName());
//        System.out.println("####################### added case: "+ addedCase.getName());

        ICaseBase cb = this.project.getCB("Example_CB");
        cb.addCase(instance);

        return null;
    }

    private Instance createInstanceFromQuery(Concept concept, HashMap<String, Object> queryContent){
        Instance newCase = new Instance(concept, ("Case #" + caseNr));
        caseNr += 1;

        try {
            for (Map.Entry<String, Object> att : queryContent.entrySet()) {
                String name = att.getKey();
                AttributeDesc attdesc = concept.getAttributeDesc(name);
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("FloatDesc")){
                    newCase.addAttribute(attdesc, Float.parseFloat(att.getValue().toString()));
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("IntegerDesc")){
                    newCase.addAttribute(attdesc, Integer.parseInt(att.getValue().toString()));
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("DoubleDesc")){
                    newCase.addAttribute(attdesc, Double.parseDouble(att.getValue().toString()));
                }
                if (attdesc.getClass().getSimpleName().equalsIgnoreCase("SymbolDesc")){
                    SymbolDesc aSymbolAtt = (SymbolDesc) attdesc;
                    if (!aSymbolAtt.isMultiple()) {
                        newCase.addAttribute(attdesc, (String) att.getValue());
                    }
                    else {
                        LinkedList<de.dfki.mycbr.core.casebase.Attribute> llAtts = new LinkedList<de.dfki.mycbr.core.casebase.Attribute>();
                        StringTokenizer st = new StringTokenizer((String) att.getValue(), ",");
                        while (st.hasMoreElements()) {
                            String symbolName = st.nextElement().toString().trim();
                            llAtts.add(aSymbolAtt.getAttribute(symbolName));
                        }

                        MultipleAttribute<SymbolDesc> muliSymbol = new MultipleAttribute<SymbolDesc>(aSymbolAtt, llAtts);
                        newCase.addAttribute(attdesc, muliSymbol);
                    }
                }
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }

        return newCase;
    }

    private String formatted(List<String> exercises){
        String formattedString = "";
        for(String exercise: exercises){
            formattedString += exercise + ";";
        }
        return formattedString;
    }

    private String crossover(String exerciseList1, String exerciseList2) {

        ArrayList<String> newExerciseList = new ArrayList<String>();
        String newExercises = "";
        String newExercises2 = "";

        //Map exerciselists to GA Chromosome
        HashMap<String, List<String>> chromosome1 = createGAChromosome(exerciseList1);
        HashMap<String, List<String>> chromosome2 = createGAChromosome(exerciseList2);
        Random generator = new Random();


        for(String key:chromosome1.keySet()) {
            int randomNum = generator.nextInt(2);
            if(randomNum < 1) {
                newExercises += formatted(chromosome1.get(key));
                newExercises2 += formatted(chromosome2.get(key));
            } else {
                newExercises += formatted(chromosome2.get(key));
                newExercises2 += formatted(chromosome1.get(key));
            }
        }

        /*
        ArrayList<String> exercisesCase1 = new ArrayList<String>(Arrays.asList(exerciseList1.split(";")));
        ArrayList<String> exercisesCase2 = new ArrayList<String>(Arrays.asList(exerciseList2.split(";")));

        //Add similar cases to newExerciseList

        Iterator<String> iter = exercisesCase1.iterator();
        while (iter.hasNext()) {
            String exercise = iter.next();

            if (exercisesCase2.contains(exercise)){
                newExerciseList.add(exercise);
                iter.remove();
                exercisesCase2.remove(exercise);
            }
        }


        //Cross add the exercises who are not similar
        System.out.println("Ex1: " + exercisesCase1);
        System.out.println("Ex2: " + exercisesCase2);
        int index1 = 0;
        int index2 = 0;
        for (int i = 0; i < exercisesCase1.size() ; i++) {
            if(exercisesCase1.size()/2.0 > i ) {
                newExerciseList.add(exercisesCase1.get(index1));
                index1 += 1;
            }
            else {
                newExerciseList.add(exercisesCase2.get(index2));
                index2 += 1;
            }

        }

        //Remove possible null value
        if(newExerciseList.contains(null)){ newExerciseList.remove(null); }

        //Create string of exercises in right format
        for (String exercise: newExerciseList) {
            newExercises += exercise+";";
        }
        */

        return newExercises;
    }


    private HashMap<String, List<String>> createGAChromosome(String exerciseList){

        HashMap<String, List<String>> gaChromosome = new HashMap<>();

        for(String exerciseId:exerciseList.split(";")){
            Instance exercise = exercise_cb.containsCase(exerciseId);
            String exerciseDescription = exercise.getAttForDesc(exerciseConcept.getAttributeDesc("Description")).toString();
            if(!gaChromosome.containsKey(exerciseDescription)){
                gaChromosome.put(exerciseDescription, new ArrayList<>());
            }
            gaChromosome.get(exerciseDescription).add(exerciseId);
        }

        return gaChromosome;
    }

    private String fitnessFunction(String closestCase, String crossover1, String crossover2){

        return null;
    }


    /*
    Mutation function.
    Happens with a probability of 1.5%.
    Mutates one random exercise, up or down one level.
    Input: ExerciseList
    Output: mutated exerciseList

    * */
    public String mutation(String exList){
        String[] exercises = exList.split(";");
        boolean flex = true;
        Random generator = new Random();
        int max = exercises.length;
        int randomNum;
        Instance exercise;

        while(flex) {
            randomNum = generator.nextInt(max-1);
            System.out.println("Mutate exercise: " + exercises[randomNum]);
            exercise = exercise_cb.containsCase(exercises[randomNum]);

            AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");

            if(!exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Flexibility")){
                flex = false;
                try {
                    Instance newExercise = changeExercise(exercise);
                    exercises[randomNum] = newExercise.getName();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }


        return String.join(";", exercises);
    }

    private Instance changeExercise(Instance exercise) throws Exception {

        AttributeDesc repDesc = exerciseConcept.getAttributeDesc("Repetitions");
        AttributeDesc levelDesc = exerciseConcept.getAttributeDesc("Level");

        Instance copied = exerciseConcept.copyInstance(exercise, ("Exercise #" + exerciseNr));
        exerciseNr += 1;

        Attribute rep = copied.getAttForDesc(repDesc);

        int oldRep = Integer.parseInt(rep.getValueAsString());
        int oldLevel = Integer.parseInt(copied.getAttForDesc(levelDesc).getValueAsString());
        String newRep = "0";

        if(canBeIncreased(exercise)) {
            System.out.println("-------------OPPJUSTER---------------");
            // Finn ut hvordan reps skal endres
            if (oldRep == 8) {
                newRep = "10";
            } else if (oldRep == 10) {
                newRep = "12";
            } else if (oldRep == 12) {
                newRep = "15";
            } else if (oldRep == 15 && canBeIncreased(exercise)) {
                newRep = "8";
            }

            // Sett nye reps
            if (newRep != "0") {
                copied.addAttribute(repDesc.getName(), repDesc.getAttribute(newRep));

                // Hvis rep 15->8 øk ex level
                if (newRep == "8") {
                    String newLevel = Integer.toString(oldLevel + 1);
                    copied.addAttribute(levelDesc, newLevel);
                }
                copied.setAdapted(true);
            }

        }
        return copied;
    }

    private boolean canBeIncreased(Instance exercise){
        AttributeDesc levelDesc = exerciseConcept.getAttributeDesc("Level");
        AttributeDesc descriptionDesc = exerciseConcept.getAttributeDesc("Description");
        AttributeDesc repDesc = exerciseConcept.getAttributeDesc("Repetitions");

        if(Integer.parseInt(exercise.getAttForDesc(repDesc).getValueAsString()) < 15){
            return true;
        }

        if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the abdominal muscles")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 6);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the back extensors")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 4);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the gluts and back extensors")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 4);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength exercises for the oblique abdominals and rotator muscle in the back")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 6);
        } else if(exercise.getAttForDesc(descriptionDesc).getValueAsString().equals("Strength “Strong in mid-position” ")){
            return (Integer.parseInt(exercise.getAttForDesc(levelDesc).getValueAsString()) < 4);
        }

        return false;
    }

}
